// Title:   Diffuser_DialogImpl.h
// Created: Thu Sep 25 15:11:43 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __DIFFUSER_DIALOGIMPL_H_
#define __DIFFUSER_DIALOGIMPL_H_

#include "../UI/Diffuser_Dialog.h"

#include "VolGrid.h"
#include "VolGridReader.h"
#include "VolGridParticleSim.h"
#include "VolGridEllipticalMC.h"
#include "VolGridMesh.h"
#include "BilateralDenoiser.h"
#include "MLSApplication.h"
#include "Edge.h"

#include <qwidget.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <qapplication.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qslider.h>
#include <qlabel.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qdatetime.h>
#include <qspinbox.h>
#include <qtabwidget.h>
#include <qfiledialog.h>
#include <qtextstream.h>
#include <qfile.h>
#include <qmessagebox.h>

#include <set>
#include <vector>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/DataStructures/src/AdjacencyGraph.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"


/**
 *
 * @author Sascha Scandella
 * @version 1.2
 */
class Diffuser_DialogImpl : public Diffuser_DialogBase {

    Q_OBJECT

public:

	/**
	 * This method creates an instance of this <code>Diffuser_DialogImpl</code>, if necessary,
	 * and returns it. Only one instance at a time exists (singleton pattern).
	 *
	 * @return an instance of this <code>Diffuser_DialogImpl</code>
	 */
	static Diffuser_DialogImpl *getInstance();

	/**
	 * Destroys the instance of this <code>Diffuser_DialogImpl</code>, if there
	 * is one.
	 */
	static void destroyInstance();

	/// Reference to the volumetric grid.
	VolGrid				       *grid;
	/// Reference to the narrow band.
	VolGridNarrowBandIterator  *narrowBand;
	/// Reference to the actual pixmap.
	QPixmap					   *actualPixmap;
	/// Maximal depth
	int					       maxDepth;
	/// Indicates which slice is active
	PLANE					   plane;
	/// Reference to <i>Particle Simulation</i>
	VolGridParticleSimulation  *particleSim;
	/// Reference to reader
	VolGridReader			   *reader;
	/// Distance field calculation method
	DISTANCEFIELDCALC		   distanceFieldCalcMethod;
	/// <i>Moving Least Squares</i> polynomial size
	int							polynomial;

private:

	Diffuser_DialogImpl (QWidget *parent = 0, const char *name = 0, bool modal = false, WFlags flags = 0);

protected:
	
	// To suppress warnings...
	virtual ~Diffuser_DialogImpl();

private:

	// initialises this dialog, setting all text labels, connects the signals
	void initDialog();

	/**
	 * This method creates a slice image.
	 *
	 * @param grid
	 *				is the current volumetric grid.
	 * @param plane
	 *				specifies the desired plane.
	 * @param sliceNumber
	 *				specifies the desired slice number.
	 * @param level
	 *				defines the depth level of the octree.
	 *
	 * @return returns a pointer to a <code>QPixmap</code>.
	 */
	QPixmap *createSliceImage(VolGrid *grid, PLANE plane, int sliceNumber, int level);

	/**
	 * Updates the slice picture.
	 *
	 * @param sliceNumber
	 *				specifies the slice number.
	 * @param plane
	 *				specifies the plane.
	 */
	void updateSlicePicture(int sliceNumber, PLANE plane);

	/**
	 * Calculates the coordinates of the bounding box.
	 *
	 * @param minimum
	 *				is the bottom left corner.
	 * @param maximum
	 *				is the upper right corner.
	 */
	void getBoundingBox(Vector3D &minimum, Vector3D &maximum);

	/**
	 * Get the filter values.
	 */
	void getFilterValues(VolGrid *grid);

private slots:
	
	// Preprocessing
	void handle_BilateralSmoothing();
	void handle_SmoothNormals();
	void handle_RemoveCorruptSurfels();
	void handle_ClassifyBadNormals();
	void handle_ToleranceValueSliderChanged(int value);
	void handle_ToleranceValueLineEditChanged(const QString& string);
	void handle_FlipSelectedNormals();
	void handle_RemoveBadNormals();
	void handle_CenterOfMass();
	void handle_CenterOfMassRemoveSelected();
	void handle_FlipAllNormals();

	// Distance Field Calculation
	void handle_CalculationMethod(const QString&);
	void handle_ResolutionChanged(const QString &string);

	// Volumetric Diffusion
	void handle_CalculateGrid();
	void handle_ClearGrid();
	void handle_FilterChanged(const QString&);
	void handle_Smoother();
	void handle_FindBoundaryNodes();
	void handle_ClearBoundaryNodes();
	void handle_FillHoles();
	void handle_ExpandOctree();

	// Surface Reconstruction
	void handle_GenerateSurfelSurface();
	void handle_InitParticleSimulation();
	void handle_Repel();	
	void handle_GenerateMesh();
	void handle_MLS();
	void handle_PolynomialChanged(const QString &string);

	// General
	void handle_XYPlane();
	void handle_YZPlane();
	void handle_XZPlane();
	void handle_SliderValueChanged(int value);
	void handle_DisableHoleMarker();
	void handle_WriteFile();	// writes points with normals
	void handle_ReadFile();		// reads points with normals

	// Debug
	void handle_LoadGrid();		// only for debugging
	void handle_SaveGrid();		// only for debugging
	void handle_WriteStatistics();  // writes statistical data

};

#endif  // __DIFFUSER_DIALOGIMPL_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
