// Title:   OcTree.cpp
// Created: Thu Sep 25 15:10:46 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "OcTree.h"

void OcTree::init(int maxNumberOfLevels) {

	// initializes some values used for localization of neighbours
	this->maxActualDepth = 0; 
	this->nLevels = maxNumberOfLevels;
	this->rootLevel = this->nLevels - 1;
	this->maxVal = pow(2, this->rootLevel);
}


void OcTree::copyTraversalNode(OcTreeTraversalNode *destination, OcTreeTraversalNode *source) {

	// Copy the traversal node data.
    destination->center = source->center;
    destination->length = source->length;
    destination->depth = source->depth;
    destination->node = source->node;
}


inline bool OcTree::isPointInside(const Vector3D &point) {
	
	for (int i = 0; i < 3; i++) {
		if ((point[i] > 1.0f) || (point[i] < 0.0f))
			return false;
	}

	return true;
}


bool OcTree::getCellForPoint(const Vector3D &point, OcTreeTraversalNode *travNode) {

    // Determine the x, y and z locational codes of the point's position.
    unsigned int xLocCode = (unsigned int) (point[0] * this->maxVal);
    unsigned int yLocCode = (unsigned int) (point[1] * this->maxVal);
	unsigned int zLocCode = (unsigned int) (point[2] * this->maxVal);

	if (isPointInside(point)) {
		// Follow the branching patterns of the locational codes from the root cell
		// to locate the leaf cell containing point
		travNode->node = this->root;		
		unsigned int nextLevel = this->rootLevel - 1;
		OT_TRAVERSE(travNode->node, nextLevel, xLocCode, yLocCode, zLocCode);
		travNode->node->level = nextLevel + 1;
		return true;
	}
	else
		return false;
}


bool OcTree::traverseNode(OcTreeTraversalNode *travNode, OcTreeIterator *iter) {

    int i;

    // Only iterate if it is a leaf node or if showInternalNodes is set.
    if (travNode->node->children == NULL || iter->showInternalNodes) {
        if (!iter->iterate(travNode))
            return false;
    }

    // Return true for a leaf.
	if (travNode->node->children == NULL)
        return true;

    // Recursively traverse all children.
    OcTreeTraversalNode *next = nodeFactory->newTraversalNode();

	// Adjust depth and length.
	next->depth = travNode->depth + 1;
	next->length = 0.5 * travNode->length;

    for (i = 0; i < 8; i++) {
    	next->node = travNode->node->children[i];
        if (next->node == NULL)
            continue;

        next->center = travNode->center + next->length * octreeChildCenterOffset[i];
        this->initTraversalNode(next, travNode, i);
        if (!this->traverseNode(next, iter))
            return false;
    }

	// Free the traversal node.
    nodeFactory->freeTraversalNode(next);

    return true;
}


bool OcTree::traverse(OcTreeIterator *iter) {

	bool ret;
    OcTreeTraversalNode *travNode;

	travNode = nodeFactory->newTraversalNode();
    travNode->center = this->center;
    travNode->length = this->length;
    travNode->depth = 0;
    travNode->node = root;
    this->initTraversalNode(travNode, travNode, -1);
	
	// Traversal.
	ret = this->traverseNode(travNode, iter);
    nodeFactory->freeTraversalNode(travNode);

    return ret;
}


bool OcTreeNode::getNeighbourCellsTest(OcTree *ocTree) {
/*
	int i, j;
	OcTreeNode *node1, *node2;

	// 1st test:
	//
	// Tested routines: getNeighbourLeft(ocTree)
	//					getNeighbourRight(ocTree)
	//					getNeighbourFront(ocTree)
	//					getNeighbourRear(ocTree)
	//					getNeighbourBottom(ocTree)
	//					getNeighbourTop(ocTree)
	// 
	// If a method returns a neighbour the node is tested with the counterpart
	// of the first method (e.g. First call getNeighbourLeft(ocTree), then switch
	// to the left node and call getNeighbourRight(ocTree). Afterwards test if nodes
	// are identical).
	// 
	// The test only returns false if the nodes are not identical.
	
	node1 = this->getNeighbourLeft(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourRight(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the left side.
	}

	node1 = this->getNeighbourRight(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourLeft(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the right side.
	}

	node1 = this->getNeighbourFront(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourRear(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the front side.
	}

	node1 = this->getNeighbourRear(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourFront(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the rear side.
	}

	node1 = this->getNeighbourBottom(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourTop(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the bottom side.
	}

	node1 = this->getNeighbourTop(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourBottom(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else if (this->level != node1->level)
		{ } // different nodes because not on the same level
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the top side.
	}

	// 2nd test:
	//
	// getNeighbourCells is tested for all possible neighbours.
	// The testing method is the same as above.
	
	this->getNeighbourCells(ocTree);

	i = 0;
	for (j = 0; j < 26; j++) {
		if ((this->neighbours[j]) != NULL) {
			this->neighbours[j]->getNeighbourCells(ocTree);
			
			if ((this->neighbours[j]->neighbours[-j + 25]) == (this)) {
				// same nodes
			}
			else if (this->level != this->neighbours[j]->level) {
				// qDebug("Different levels: %i - %i", this->level, this->neighbours[j]->level);
			}
			else {
				qDebug("*** Warning! Nodes are not identical.");
				return false;
			}
			i++;
		}
	}
*/
	return true;
}


bool OcTreeNode::getNeighbourCellsSameLevelTest(OcTree *ocTree) {
/*
	int i, j;
	OcTreeNode *node1, *node2;

	// 1st test:
	//
	// Tested routines: getNeighbourLeftSameLevel(ocTree)
	//					getNeighbourRightSameLevel(ocTree)
	//					getNeighbourFrontSameLevel(ocTree)
	//					getNeighbourRearSameLevel(ocTree)
	//					getNeighbourBottomSameLevel(ocTree)
	//					getNeighbourTopSameLevel(ocTree)
	// 
	// If a method returns a neighbour the node is tested with the counterpart
	// of the first method (e.g. First call getNeighbourLeftSameLevel(ocTree), then switch
	// to the left node and call getNeighbourRightSameLevel(ocTree). Afterwards test if nodes
	// are identical).
	// 
	// The test only returns false if the nodes are not identical.
	
	node1 = this->getNeighbourLeftSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourRightSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the left side.
	}

	node1 = this->getNeighbourRightSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourLeftSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the right side.
	}

	node1 = this->getNeighbourFrontSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourRearSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the front side.
	}

	node1 = this->getNeighbourRearSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourFrontSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the rear side.
	}

	node1 = this->getNeighbourBottomSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourTopSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the bottom side.
	}

	node1 = this->getNeighbourTopSameLevel(ocTree);
	if (node1 != NULL) {
		node2 = node1->getNeighbourBottomSameLevel(ocTree);
		if ((node2 != NULL) && (node2 == this))
		{ } // same nodes
		else {
			qDebug("*** Warning! Nodes are not identical.");
			return false;
		}
	}
	else {
		// No node on the top side.
	}

	// 2nd test:
	//
	// getNeighbourCellsSameLevel is tested for all possible neighbours.
	// The testing method is the same as above.
	
	this->getNeighbourCellsSameLevel(ocTree);

	i = 0;
	for (j = 0; j < 26; j++) {
		if ((this->neighbours[j]) != NULL) {
			this->neighbours[j]->getNeighbourCellsSameLevel(ocTree);
			
			if ((this->neighbours[j]->neighbours[-j + 25]) == (this)) {
				// same nodes
			}
			else {
				qDebug("*** Warning! Nodes are not identical.");
				return false;
			}
			i++;
		}
	}
*/
	return true;
}


void OcTreeNode::getCoordinates(Vector3D &point, OcTree *ocTree) {

	int i;
	OcTreeNode *currentNode;

	// Clear point coordinates.
	point[0] = point[1] = point[2] = 0;
	
	unsigned int nextLevel = ocTree->rootLevel - 1;
	currentNode = ocTree->root;
	i = 1;

	// Write coordinates (according to the locational codes) to the vector point.
    while (currentNode->children) {
        unsigned int childBranchBit = 1 << (nextLevel);
        unsigned int childIndex = ((((this->xLocCode) & childBranchBit) >> (nextLevel))
					+ (((this->yLocCode) & childBranchBit) >> ((nextLevel - 1)))
					+ (((this->zLocCode) & childBranchBit) >> ((nextLevel - 2))));
		if (((this->xLocCode) & childBranchBit)) { point[0] += pow(0.5, i); }
		if (((this->yLocCode) & childBranchBit)) { point[1] += pow(0.5, i); }
		if (((this->zLocCode) & childBranchBit)) { point[2] += pow(0.5, i); }
        nextLevel--;
		i++;
		currentNode = (currentNode->children[childIndex]);
	}
}


OcTreeNode *OcTreeNode::getNeighbourLeft(OcTree *ocTree) {
    
	// No left neighbour if this is the left side of the octree.
	if (this->xLocCode == 0)
		return NULL;
	else {
	    
		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible left neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int xLeftLocCode = this->xLocCode - 0x00000001;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible left neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = xLocCode ^ xLeftLocCode;
		OcTreeNode *left = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(left, nextLevel, diff);
		
		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest left
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(left,nextLevel,xLeftLocCode,yLocCode,zLocCode,nodeLevel);
		return left;
	}
}


OcTreeNode *OcTreeNode::getNeighbourLeftSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourLeft(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode *OcTreeNode::getNeighbourRight(OcTree *ocTree) {

	unsigned int binaryCellSize = 1 << this->level;

	// No right neighbour if this is the right side of the octree.	
	if ((this->xLocCode + binaryCellSize) >= (1 << ocTree->rootLevel))
		return NULL;
	else {

		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible right neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int xRightLocCode = this->xLocCode + binaryCellSize;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible right neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = xLocCode ^ xRightLocCode;
		OcTreeNode *right = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(right, nextLevel, diff);

		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest right
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(right,nextLevel,xRightLocCode,yLocCode,zLocCode,nodeLevel);
		return right;
	}
}


OcTreeNode *OcTreeNode::getNeighbourRightSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourRight(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode *OcTreeNode::getNeighbourFront(OcTree *ocTree) {
    
	// No front neighbour if this is the front side of the octree.
	if (this->yLocCode == 0)
		return NULL;
	else {
	    
		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible front neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int yFrontLocCode = this->yLocCode - 0x00000001;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible front neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = yLocCode ^ yFrontLocCode;
		OcTreeNode *front = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(front, nextLevel, diff);
		
		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest front
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(front,nextLevel,xLocCode,yFrontLocCode,zLocCode,nodeLevel);
		return front;
	}
}


OcTreeNode *OcTreeNode::getNeighbourFrontSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourFront(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode *OcTreeNode::getNeighbourRear(OcTree *ocTree) {

	unsigned int binaryCellSize = 1 << this->level;

	// No rear neighbour if this is the rear side of the octree.	
	if ((this->yLocCode + binaryCellSize) >= (1 << ocTree->rootLevel))
		return NULL;
	else {

		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible rear neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int yRearLocCode = this->yLocCode + binaryCellSize;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible rear neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = yLocCode ^ yRearLocCode;
		OcTreeNode *rear = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(rear, nextLevel, diff);

		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest rear
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(rear,nextLevel,xLocCode,yRearLocCode,zLocCode,nodeLevel);
		return rear;
	}
}


OcTreeNode *OcTreeNode::getNeighbourRearSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourRear(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode *OcTreeNode::getNeighbourBottom(OcTree *ocTree) {
    
	// No bottom neighbour if this is the bottom side of the octree.
	if (this->zLocCode == 0)
		return NULL;
	else {
	    
		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible bottom neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int zBottomLocCode = this->zLocCode - 0x00000001;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible bottom neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = zLocCode ^ zBottomLocCode;
		OcTreeNode *bottom = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(bottom, nextLevel, diff);
		
		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest bottom
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(bottom,nextLevel,xLocCode,yLocCode,zBottomLocCode,nodeLevel);
		return bottom;
	}
}


OcTreeNode *OcTreeNode::getNeighbourBottomSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourBottom(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode *OcTreeNode::getNeighbourTop(OcTree *ocTree) {

	unsigned int binaryCellSize = 1 << this->level;

	// No top neighbour if this is the top side of the octree.	
	if ((this->zLocCode + binaryCellSize) >= (1 << ocTree->rootLevel))
		return NULL;
	else {

		// Get node's x, y and z locational codes and the x locational code of the
        // node's smallest possible top neighbour.
		unsigned int xLocCode = this->xLocCode;
		unsigned int yLocCode = this->yLocCode;
		unsigned int zLocCode = this->zLocCode;
		unsigned int zTopLocCode = this->zLocCode + binaryCellSize;

		// Determine the smallest common ancestor of the node and the node's
        // smallest possible top neighbour.
		unsigned int nodeLevel, nextLevel;
		unsigned int diff = zLocCode ^ zTopLocCode;
		OcTreeNode *top = this;
		nodeLevel = nextLevel = this->level;
		OT_GET_COMMON_ANCESTOR(top, nextLevel, diff);

		// Start from the smallest common ancestor and follow the branching
        // patterns of the locational codes downward to the smallest top
        // neighbour of size greater than or equal to node.
		nextLevel--;
		OT_TRAVERSE_TO_LEVEL(top,nextLevel,xLocCode,yLocCode,zTopLocCode,nodeLevel);
		return top;
	}
}


OcTreeNode *OcTreeNode::getNeighbourTopSameLevel(OcTree *ocTree) {
	
	OcTreeNode *node = this->getNeighbourTop(ocTree);

	if (node != NULL && (node->level != this->level))
			node = NULL;

	return node;
}


OcTreeNode **OcTreeNode::getNeighbourCells(OcTree *ocTree) {

	int i;

	// Init neighbours array.
	OcTreeNode **neighbours = new OcTreeNode *[26];
	
	// There are no right neighbours if this is the right side of the octree and
	// no bottom neighbors if this is the bottom of the octree, etc...
	unsigned int binCellSize = 1 << this->level;
	unsigned int noLeft = (this->xLocCode == 0) ? 1 : 0;
    unsigned int noRight = ((this->xLocCode + binCellSize) >= (1 << ocTree->rootLevel)) ? 1 : 0;
   	unsigned int noFront = (this->yLocCode == 0) ? 1 : 0;
	unsigned int noRear = ((this->yLocCode + binCellSize) >= (1 << ocTree->rootLevel)) ? 1 : 0;
	unsigned int noBottom = (this->zLocCode == 0) ? 1 : 0;
	unsigned int noTop = ((this->zLocCode + binCellSize) >= (1 << ocTree->rootLevel)) ? 1 : 0;

	// calculate all locational codes
	unsigned int xLocCode = this->xLocCode;
	unsigned int yLocCode = this->yLocCode;
	unsigned int zLocCode = this->zLocCode;
	unsigned int xLeftLocCode = this->xLocCode - 0x00000001;
    unsigned int xRightLocCode = this->xLocCode + binCellSize;
	unsigned int yFrontLocCode = this->yLocCode - 0x00000001;
    unsigned int yRearLocCode =  this->yLocCode + binCellSize;
	unsigned int zBottomLocCode = this->zLocCode - 0x00000001;
    unsigned int zTopLocCode = this->zLocCode + binCellSize;

	unsigned int nodeLevel, leftLevel, rightLevel, frontLevel, rearLevel, bottomLevel, topLevel;
	nodeLevel = leftLevel = rightLevel = frontLevel = rearLevel = bottomLevel = topLevel = this->level;
	unsigned int diff;

	// Pointer to store common ancestors.
	OcTreeNode* commonLeft;
	OcTreeNode* commonRight;
	OcTreeNode* commonFront;
	OcTreeNode* commonRear;
	OcTreeNode* commonBottom;
	OcTreeNode* commonTop;

	// There are 26 possible neighbours to locate. If a corresponding neighbour does
	// not exist then references are temporarily set to *this and at the end to NULL.
    if (noLeft) {
		neighbours[0] = this;
		neighbours[3] = this;
		neighbours[6] = this;
		neighbours[9] = this;
		neighbours[12] = this;
		neighbours[14] = this;
		neighbours[17] = this;
		neighbours[20] = this;
		neighbours[23] = this;
	}
	else {
		// calculate neighbour 12
		unsigned int level = this->level;
		diff = xLocCode ^ xLeftLocCode;
		commonLeft = this;
		OT_GET_COMMON_ANCESTOR(commonLeft, level, diff);
		leftLevel = level;
		neighbours[12] = commonLeft;
		level--;
		OT_TRAVERSE_TO_LEVEL(neighbours[12],level,xLeftLocCode,yLocCode,zLocCode,nodeLevel);
	}

	
    if (noRight) {
		neighbours[2] = this;
		neighbours[5] = this;
		neighbours[8] = this;
		neighbours[11] = this;
		neighbours[13] = this;
		neighbours[16] = this;
		neighbours[19] = this;
		neighbours[22] = this;
		neighbours[25] = this;
	}
	else {
		// calculate neighbour 13
		unsigned int level = this->level;
		diff = xLocCode ^ xRightLocCode;
		commonRight = this;
		OT_GET_COMMON_ANCESTOR(commonRight, level, diff);
		rightLevel = level;
		neighbours[13] = commonRight;
		level--;
		OT_TRAVERSE_TO_LEVEL(neighbours[13],level,xRightLocCode,yLocCode, zLocCode,nodeLevel);
	}
	if (noFront) {
		neighbours[0] = this;
		neighbours[1] = this;
		neighbours[2] = this;
		neighbours[9] = this;
		neighbours[10] = this;
		neighbours[11] = this;
		neighbours[17] = this;
		neighbours[18] = this;
		neighbours[19] = this;
	}
	else {
		// calculate neighbour 10
		if (neighbours[10] != this) {
			unsigned int level = this->level;
			diff = yLocCode ^ yFrontLocCode;
			commonFront = this;
			OT_GET_COMMON_ANCESTOR(commonFront, level, diff);
			frontLevel = level;
			neighbours[10] = commonFront;
			level--;
			OT_TRAVERSE_TO_LEVEL(neighbours[10],level,xLocCode,yFrontLocCode,zLocCode,nodeLevel);
		}

		// calculate neighbour 9
		if (neighbours[9] != this) {
			if (leftLevel >= frontLevel) {
				unsigned int level = leftLevel - 1;
				neighbours[9] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[9],level,xLeftLocCode,yFrontLocCode,zLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[9] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[9],level,xLeftLocCode,yFrontLocCode,zLocCode,nodeLevel);
			}
		}
		// calculate neighbour 11
		if (neighbours[11] != this) {
			if (rightLevel >= frontLevel) {
				unsigned int level = rightLevel - 1;
				neighbours[11] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[11],level,xRightLocCode,yFrontLocCode,zLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[11] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[11],level,xRightLocCode,yFrontLocCode,zLocCode,nodeLevel);
			}
		}
	}
    if (noRear) {
		neighbours[6] = this;
		neighbours[7] = this;
		neighbours[8] = this;
		neighbours[14] = this;
		neighbours[15] = this;
		neighbours[16] = this;
		neighbours[23] = this;
		neighbours[24] = this;
		neighbours[25] = this;
	}
	else {
		// calculate neighbour 15
		if (neighbours[15] != this) {
			unsigned int level = this->level;
			diff = yLocCode ^ yRearLocCode;
			commonRear = this;
			OT_GET_COMMON_ANCESTOR(commonRear, level, diff);
			rearLevel = level;
			neighbours[15] = commonRear;
			level--;
			OT_TRAVERSE_TO_LEVEL(neighbours[15],level,xLocCode,yRearLocCode,zLocCode,nodeLevel);

		}
		// calculate neighbour 14
		if (neighbours[14] != this) {
			if (leftLevel >= rearLevel) {
				unsigned int level = leftLevel - 1;
				neighbours[14] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[14],level,xLeftLocCode,yRearLocCode,zLocCode,nodeLevel);
			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[14] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[14],level,xLeftLocCode,yRearLocCode,zLocCode,nodeLevel);
			}
		}
		// calculate neighbour 16
		if (neighbours[16] != this) {
			if (rightLevel >= rearLevel) {
				unsigned int level = rightLevel - 1;
				neighbours[16] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[16],level,xRightLocCode,yRearLocCode,zLocCode,nodeLevel);

			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[16] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[16],level,xRightLocCode,yRearLocCode,zLocCode,nodeLevel);
			}
		}
	}
    if (noBottom) {
		for (int j = 0; j < 9; j++)
			neighbours[j] = this;
	}
	else {
		// calculate neighbour 4
		if (neighbours[4] != this) {
			unsigned int level = this->level;
			diff = zLocCode ^ zBottomLocCode;
			commonBottom = this;
			OT_GET_COMMON_ANCESTOR(commonBottom, level, diff);
			bottomLevel = level;
			neighbours[4] = commonBottom;
			level--;
			OT_TRAVERSE_TO_LEVEL(neighbours[4],level,xLocCode,yLocCode,zBottomLocCode,nodeLevel);
		}
		// calculate neighbour 1
		if (neighbours[1] != this) {
			if (frontLevel > bottomLevel) {
				unsigned int level = frontLevel - 1;
				neighbours[1] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[1],level,xLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = bottomLevel - 1;
				neighbours[1] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[1],level,xLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 3
		if (neighbours[3] != this) {
			if (leftLevel > bottomLevel) {
				unsigned int level = leftLevel - 1;
				neighbours[3] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[3],level,xLeftLocCode,yLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = bottomLevel - 1;
				neighbours[3] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[3],level,xLeftLocCode,yLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 5
		if (neighbours[5] != this) {
			if (rightLevel > bottomLevel) {
				unsigned int level = rightLevel - 1;
				neighbours[5] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[5],level,xRightLocCode,yLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = bottomLevel - 1;
				neighbours[5] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[5],level,xRightLocCode,yLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 7
		if (neighbours[7] != this) {
			if (rearLevel > bottomLevel) {
				unsigned int level = rearLevel - 1;
				neighbours[7] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[7],level,xLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = bottomLevel - 1;
				neighbours[7] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[7],level,xLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 0
		if (neighbours[0] != this) {
			if ((leftLevel > bottomLevel) && (leftLevel > frontLevel)) {
				unsigned int level = leftLevel - 1;
				neighbours[0] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[0],level,xLeftLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
			else if (bottomLevel > frontLevel) {
				unsigned int level = bottomLevel - 1;
				neighbours[0] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[0],level,xLeftLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[0] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[0],level,xLeftLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 2
		if (neighbours[2] != this) {
			if ((rightLevel > bottomLevel) && (rightLevel > frontLevel)) {
				unsigned int level = rightLevel - 1;
				neighbours[2] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[2],level,xRightLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
			else if (bottomLevel > frontLevel) {
				unsigned int level = bottomLevel - 1;
				neighbours[2] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[2],level,xRightLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[2] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[2],level,xRightLocCode,yFrontLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 6
		if (neighbours[6] != this) {
			if ((leftLevel > bottomLevel) && (leftLevel > rearLevel)) {
				unsigned int level = leftLevel - 1;
				neighbours[6] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[6],level,xLeftLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
			else if (bottomLevel > rearLevel) {
				unsigned int level = bottomLevel - 1;
				neighbours[6] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[6],level,xLeftLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[6] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[6],level,xLeftLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
		}
		// calculate neighbour 8
		if (neighbours[8] != this) {
			if ((rightLevel > bottomLevel) && (rightLevel > rearLevel)) {
				unsigned int level = rightLevel - 1;
				neighbours[8] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[8],level,xRightLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
			else if (bottomLevel > rearLevel) {
				unsigned int level = bottomLevel - 1;
				neighbours[8] = commonBottom;
				OT_TRAVERSE_TO_LEVEL(neighbours[8],level,xRightLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[8] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[8],level,xRightLocCode,yRearLocCode,zBottomLocCode,nodeLevel);
			}
		}
	}
    if (noTop) {
		for (int j = 17; j < 26; j++)
			neighbours[j] = this;
	}
	else {
		// calculate neighbour 21
		if (neighbours[21] != this) {
			unsigned int level = this->level;
			diff = zLocCode ^ zTopLocCode;
			commonTop = this;
			OT_GET_COMMON_ANCESTOR(commonTop, level, diff);
			topLevel = level;
			neighbours[21] = commonTop;
			level--;
			OT_TRAVERSE_TO_LEVEL(neighbours[21],level,xLocCode,yLocCode,zTopLocCode,nodeLevel);
		}
		// calculate neighbour 18
		if (neighbours[18] != this) {
			if (frontLevel > topLevel) {
				unsigned int level = frontLevel - 1;
				neighbours[18] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[18],level,xLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = topLevel - 1;
				neighbours[18] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[18],level,xLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 20
		if (neighbours[20] != this) {
			if (leftLevel > topLevel) {
				unsigned int level = leftLevel - 1;
				neighbours[20] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[20],level,xLeftLocCode,yLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = topLevel - 1;
				neighbours[20] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[20],level,xLeftLocCode,yLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 22
		if (neighbours[22] != this) {
			if (rightLevel > topLevel) {
				unsigned int level = rightLevel - 1;
				neighbours[22] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[22],level,xRightLocCode,yLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = topLevel - 1;
				neighbours[22] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[22],level,xRightLocCode,yLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 24
		if (neighbours[24] != this) {
			if (rearLevel > topLevel) {
				unsigned int level = rearLevel - 1;
				neighbours[24] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[24],level,xLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = topLevel - 1;
				neighbours[24] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[24],level,xLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 17
		if (neighbours[17] != this) {
			if ((leftLevel > topLevel) && (leftLevel > frontLevel)) {
				unsigned int level = leftLevel - 1;
				neighbours[17] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[17],level,xLeftLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
			else if (topLevel > frontLevel) {
				unsigned int level = topLevel - 1;
				neighbours[17] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[17],level,xLeftLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[17] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[17],level,xLeftLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 19
		if (neighbours[19] != this) {
			if ((rightLevel > topLevel) && (rightLevel > frontLevel)) {
				unsigned int level = rightLevel - 1;
				neighbours[19] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[19],level,xRightLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
			else if (topLevel > frontLevel) {
				unsigned int level = topLevel - 1;
				neighbours[19] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[19],level,xRightLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = frontLevel - 1;
				neighbours[19] = commonFront;
				OT_TRAVERSE_TO_LEVEL(neighbours[19],level,xRightLocCode,yFrontLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 23
		if (neighbours[23] != this) {
			if ((leftLevel > topLevel) && (leftLevel > rearLevel)) {
				unsigned int level = leftLevel - 1;
				neighbours[23] = commonLeft;
				OT_TRAVERSE_TO_LEVEL(neighbours[23],level,xLeftLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
			else if (topLevel > rearLevel) {
				unsigned int level = topLevel - 1;
				neighbours[23] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[23],level,xLeftLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[23] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[23],level,xLeftLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
		}
		// calculate neighbour 25
		if (neighbours[25] != this) {
			if ((rightLevel > topLevel) && (rightLevel > rearLevel)) {
				unsigned int level = rightLevel - 1;
				neighbours[25] = commonRight;
				OT_TRAVERSE_TO_LEVEL(neighbours[25],level,xRightLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
			else if (topLevel > rearLevel) {
				unsigned int level = topLevel - 1;
				neighbours[25] = commonTop;
				OT_TRAVERSE_TO_LEVEL(neighbours[25],level,xRightLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
			else {
				unsigned int level = rearLevel - 1;
				neighbours[25] = commonRear;
				OT_TRAVERSE_TO_LEVEL(neighbours[25],level,xRightLocCode,yRearLocCode,zTopLocCode,nodeLevel);
			}
		}
	}

	// Change all nodes with no neighbours to NULL.
	for (i = 0; i < 26; i++) {
		if (neighbours[i] == this)
			neighbours[i] = NULL;
	}

	return neighbours;
}


OcTreeNode **OcTreeNode::getNeighbourCellsSameLevel(OcTree *ocTree) {

	OcTreeNode **neighbours;
	neighbours = this->getNeighbourCells(ocTree);

	for (int i = 0; i < 26; i++) {
		if (neighbours[i] != NULL && (this->level != neighbours[i]->level))
			neighbours[i] = NULL;
	}

	return neighbours;
}


void OcTreeStatistics::reportStatisticsAtDepth(int depth) {

    double length = 2 * ocTree->length * pow(0.5, depth);

	qDebug("Depth: %i %f", depth, length);
    qDebug("Total internal nodes: %i", internalNodeCount[depth]);
	qDebug("Total leaf nodes: %i", leafCount[depth]);
}


void OcTreeStatistics::reportStatistics() {

	int i;
    int totalLeaf = 0;
    int totalInternal = 0;

    for (i = 0; i < ocTree->maxActualDepth + 1; i++) {
        reportStatisticsAtDepth(i);
		qDebug("");
        totalLeaf += leafCount[i];
        totalInternal += internalNodeCount[i];
    }

	qDebug("Maximum depth: %i", ocTree->maxActualDepth);
    qDebug("Total nodes: %i", totalNodes);
    qDebug("Total internal nodes: %i", totalInternal);
    qDebug("Total leaf nodes: %i", totalLeaf);

}


void OcTreeStatistics::collectStatistics(OcTree *octree) {

	int i;

    this->clear();
    this->ocTree = octree;
    leafCount = new int[this->ocTree->maxActualDepth + 1];
    internalNodeCount = new int[this->ocTree->maxActualDepth + 1];

    for(i = 0; i < this->ocTree->maxActualDepth+1; i++) {
        leafCount[i] = 0;
        internalNodeCount[i] = 0;
    }

    totalNodes = 0;
    this->ocTree->traverse(this);
}


bool OcTreeStatistics::iterate(OcTreeTraversalNode *travNode) {

	totalNodes++;
    if (travNode->node->children == NULL)
        leafCount[travNode->depth]++;
    else
        internalNodeCount[travNode->depth]++;
	return true;
}


void OcTreeStatistics::clear() {

	if (leafCount != NULL) {
        delete[] leafCount;
        leafCount = NULL;
    }
    if (internalNodeCount != NULL) {
        delete[] internalNodeCount;
        internalNodeCount = NULL;
    }
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
