// Title:   PlaneFit.h
// Created: Thu Sep 25 15:10:19 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PLANEFIT_H_
#define __PLANEFIT_H_

#include "../../../DataTypes/src/Vector3D.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include <qgl.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>

/**
 * Calculates a plane of form n[0]*x + n[1]*y + n[2]*z + d = 0 
 * in the weighted least squares sense:
 */
class PlaneFit {

public:
	
	PlaneFit();
	
   /**
	* Plane fit in the weighted least squares sense with fixed sigma.
	* @param nbHood
	*		kd-Tree
	* @param source
	*		source location
	* @param nofNeighbours
	*		number of neighbours
	* @param param
	*		sigma for the Gaussian function.
	*/	
	void doPlaneFit(NeighbourHood *nbHood, const Vector3D& source, int nofNeighbours, float param);

   /**
	* Adaptive plane fit in the weighted least squares sense.
	* @param nbHood
	*		kd-Tree
	* @param source
	*		source location
	* @param nofNeighbours
	*		number of neighbours
	*/	
	void doPlaneFitAdaptiveWeight(NeighbourHood *nbHood, const Vector3D& source, int nofNeighbours);


   /**
	* @param eigenvals
	*		copies the eigenvalues
	*/
	void getEigenValues(float eigenvals[3]) const;
	
	
   /**
	* @return
	*		the normal of the plane.
	*/
	Vector3D getNormal() const;
	
   /**
	* @return
	*		the anchor of the plane.
	*/
	Vector3D getCenter() const;
	
	float getD() const;
	
   /**
	* Resets values
	*/
	void reset();

   /**
	* @param souce
	*				source point.
	* @return
	*		the distance from source
	*/
	
	float getDistance(Vector3D source) const;
   
   /**
	* @return
	*		true if plane is valid
	*/
	bool isValid() const;

public:
	/// normal vector
	Vector3D	_normal;
	/// center
	Vector3D	_center;
	/// distance 
	float		_d;
	/// eigenvalues of the covariance matrix
	float		_eigenvalues[3];
	/// indicates wheter plane is valid
	bool		_isvalid;
};

#endif	// __PLANEFIT_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
