// Title:   VolGridEllipticalMC.cpp
// Created: Thu Sep 25 15:10:11 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "VolGridEllipticalMC.h"

void VolGridEllipticalMarchingCubes::marchingCubes(float values[8], float level, int& nVertices, int edgeStart[12], int edgeEnd[12],
									               float param[12], int& nTriangles, int triangles[5][3]) {

	// Compute the "case" number
    int type = 0;
    if (values[0] > level) type |=   1;
    if (values[1] > level) type |=   2;
    if (values[2] > level) type |=   4;
    if (values[3] > level) type |=   8;
    if (values[4] > level) type |=  16;
    if (values[5] > level) type |=  32;
    if (values[6] > level) type |=  64;
    if (values[7] > level) type |= 128;

    // Return if nothing to do
    if (type == 0 || type == 255) {
		nVertices = 0;
		nTriangles = 0;
		return;
    }

    // Get info about this case
    MCEntry *info = &(table[type]);

    // Copy info to output arguments
    nVertices  = info->nVertices;
    memcpy(edgeStart, info->edgeStart, 12 * sizeof(int));
    memcpy(edgeEnd, info->edgeEnd, 12 * sizeof(int));
    nTriangles = info->nTriangles;
    memcpy(triangles, info->triangles, 15 * sizeof(int));

    // Compute the interpolation parameter
    for (int i = 0; i < info->nVertices; i++) {
		float f0 = (float)values[info->edgeStart[i]];
		float f1 = (float)values[info->edgeEnd[i]];
		float diff = f1 - f0;
        param[i] =  diff==0 ? .5 : (level - f0) / diff;
    }
}


bool VolGridEllipticalMarchingCubes::iterate(OcTreeTraversalNode *ocTravNode) {

	int i, j;
	float xyz[8][3];

	VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// Determine if child is a leaf node at maximum depth.
	if ((travNode->depth == this->grid->getMaxDepth()) && (((VolGridNode *) (travNode->node))->nodeType == LEAF) &&
		(((VolGridNode *) (travNode->node))->valid == true)) {
		
		// Determine coordinates of the vertices.
		for (i = 0; i < 8; i++) {
			
			// set the correct x, y and z coordinates
			xyz[i][0] = travNode->center[0] + travNode->length * octreeChildCenterOffsetInt[i][0];
			xyz[i][1] = travNode->center[1] + travNode->length * octreeChildCenterOffsetInt[i][1];
			xyz[i][2] = travNode->center[2] + travNode->length * octreeChildCenterOffsetInt[i][2];

		}

		// Call marching cubes algo for current cell
		int nVertices;
		int edgeStart[12], edgeEnd[12];
		float param[12];
		int nTriangles;
		int triangles[5][3];
		float f[8];
		
		for (j = 0; j < 8; j++) {
			f[j] = *(node->distanceValues[j]);
		}
		
		this->marchingCubes(f, this->isoValue, nVertices, edgeStart, edgeEnd, param, nTriangles, triangles);

		// Interpolate surface vertices
		float vertex[12][3];
		for (int v = 0; v < nVertices; v++) {
			int index0 = edgeStart[v];
			int index1 = edgeEnd[v];
			float t = param[v];
			vertex[v][0] = (1 - t) * xyz[index0][0] + t * xyz[index1][0];
			vertex[v][1] = (1 - t) * xyz[index0][1] + t * xyz[index1][1];
			vertex[v][2] = (1 - t) * xyz[index0][2] + t * xyz[index1][2];
		}

		// calculate 
		for (int t = 0; t < nTriangles; t++) {
		
			int a = triangles[t][0];
			int b = triangles[t][1];
			int c = triangles[t][2];
			float area = 0.0;
			float radius = 0.0;
			
			Vector3D vertex1, vertex2, vertex3, edge1, edge2;
			Vector3D normal, centroid;
			// the three points projected on the plane
			Vector2D p0, p1, p2;
			// u and v and normal are mutually perpendicular
			Vector3D u, v;

			for (int i = 0; i < 3; i++) {
				vertex1[i] = vertex[a][i];
				vertex2[i] = vertex[b][i];
				vertex3[i] = vertex[c][i];
				// calculate the centroid of the triangle (this is the same as the center of the steiner circum ellipse)
				centroid[i] = (vertex[a][i] + vertex[b][i] + vertex[c][i]) / 3;
			}

			edge1 = vertex2 - vertex1;
			edge2 = vertex3 - vertex1;

			// compute cross product to get the normal
			normal = Vector3D::crossProduct(edge2, edge1);
			normal.normalize();

			// u and v are choosen, so that u, v and normal are mutually perpendicular
			// set u[0] = 1 and u[2] = 1
			if (normal[2] == 0) {
				u[0] = 0;
				u[1] = 0;
				u[2] = 1;
			}
			else {
				u[0] = 1;
				u[1] = 1;
				u[2] = -(normal[0] + normal[1]) / normal[2];
			}
			
			// normalize u
			u.normalize();

			// calculate v
			v = Vector3D::crossProduct(normal, u);
			v.normalize();
			
			// calculate the projected points on the plane
			p0[0] = u[0] * (vertex1[0] - centroid[0]) + u[1] * (vertex1[1] - centroid[1]) + u[2] * (vertex1[2] - centroid[2]);
			p0[1] = v[0] * (vertex1[0] - centroid[0]) + v[1] * (vertex1[1] - centroid[1]) + v[2] * (vertex1[2] - centroid[2]);

			p1[0] = u[0] * (vertex2[0] - centroid[0]) + u[1] * (vertex2[1] - centroid[1]) + u[2] * (vertex2[2] - centroid[2]);
			p1[1] = v[0] * (vertex2[0] - centroid[0]) + v[1] * (vertex2[1] - centroid[1]) + v[2] * (vertex2[2] - centroid[2]);

			p2[0] = u[0] * (vertex3[0] - centroid[0]) + u[1] * (vertex3[1] - centroid[1]) + u[2] * (vertex3[2] - centroid[2]);
			p2[1] = v[0] * (vertex3[0] - centroid[0]) + v[1] * (vertex3[1] - centroid[1]) + v[2] * (vertex3[2] - centroid[2]);

			Mgc::Matrix2 invM, M;

			// calculate the inverse of M
			invM[0][0] = (p0[0] * p0[0] + p1[0] * p1[0] + p2[0] * p2[0]) / 3;
			invM[1][0] = (p0[1] * p0[0] + p1[1] * p1[0] + p2[1] * p2[0]) / 3;
			invM[0][1] = (p0[0] * p0[1] + p1[0] * p1[1] + p2[0] * p2[1]) / 3;
			invM[1][1] = (p0[1] * p0[1] + p1[1] * p1[1] + p2[1] * p2[1]) / 3;
			
			// when using small values the eigensolver often gets a determinant of zero,
			// thus the inverse can not be determined
			M = invM.Inverse();
			
			// mathematically correct is a division by 2 - but when dividing
			// by 3 the surface quality yields better results.
			M[0][0] = M[0][0] / 3;
			M[1][0] = M[1][0] / 3;
			M[0][1] = M[0][1] / 3;
			M[1][1] = M[1][1] / 3;

			Mgc::Vector2 akEigenvector[2];

			Mgc::Real	afEigenvalue[2];

			M.EigenSolveSymmetric(afEigenvalue, akEigenvector);

			Vector2D TangentAxis12D, TangentAxis22D;
			TangentAxis12D = Vector2D(akEigenvector[0][0],akEigenvector[0][1]);
			TangentAxis22D = Vector2D(akEigenvector[1][0],akEigenvector[1][1]);

			Vector3D TangentAxis1, TangentAxis2;
	
			TangentAxis1 = Vector3D ((1/sqrt(afEigenvalue[0])) * u[0] * TangentAxis12D[0] + (1/sqrt(afEigenvalue[1])) * v[0] * TangentAxis12D[1],
									 (1/sqrt(afEigenvalue[0])) * u[1] * TangentAxis12D[0] + (1/sqrt(afEigenvalue[1])) * v[1] * TangentAxis12D[1],
									 (1/sqrt(afEigenvalue[0])) * u[2] * TangentAxis12D[0] + (1/sqrt(afEigenvalue[1])) * v[2] * TangentAxis12D[1]);
			
			TangentAxis2 = Vector3D ((1/sqrt(afEigenvalue[0])) * u[0] * TangentAxis22D[0] + (1/sqrt(afEigenvalue[1])) * v[0] * TangentAxis22D[1],
									 (1/sqrt(afEigenvalue[0])) * u[1] * TangentAxis22D[0] + (1/sqrt(afEigenvalue[1])) * v[1] * TangentAxis22D[1],
									 (1/sqrt(afEigenvalue[0])) * u[2] * TangentAxis22D[0] + (1/sqrt(afEigenvalue[1])) * v[2] * TangentAxis22D[1]);

			
			// when the vertices of a triangle are very close to each other it might
			// be that the determinant of the matrix M is not defined, thus leading
			// to tangent axes with zero length. these surfels should not be drawn.
			if ((TangentAxis1.getLength() != 0) && (TangentAxis2.getLength() != 0)) {
				this->normalList.push_back(normal);
				this->posList.push_back(centroid);
				this->tangentAxesList.push_back(TangentAxis1);
				this->tangentAxesList.push_back(TangentAxis2);
			}
		}
	}

	return true;
}


void VolGridEllipticalMarchingCubes::generateSurface() {

	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;

	int i;

	scene = Scene::getInstance();
		
	newObject = new Object("Elliptical Marching Cubes");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION;
	surfels->setPropertyDescriptor(pd);

	int nofSurfels = this->posList.size();
	surfels->reserve(this->posList.size());

	for (i = 0; i < nofSurfels; i++) {
		// extreme thin surfels could be removed
//		if (!((this->tangentAxesList.at(2*i).getLength() >=  2.5 * (this->tangentAxesList.at(2*i + 1)).getLength()) ||
//				(2.5 * this->tangentAxesList.at(2*i).getLength() <= (this->tangentAxesList.at(2*i + 1)).getLength()))) {
	  
			surfel = surfels->addSurfel(false);
			surfel->setPosition (this->posList[i]);
			surfel->setNormal(this->normalList[i]);
			surfel->setTangentAxes(this->tangentAxesList[2*i], this->tangentAxesList[2*i + 1]);
//		}
	}

	scene->addObject(newObject);
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
}


void VolGridEllipticalMarchingCubes::build(float isoValue) {

	this->grid = grid;
	this->isoValue = isoValue;

	this->grid->traverse(this);
	this->generateSurface();
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
