// Title:   VolGridReader.cpp
// Created: Thu Sep 25 15:09:58 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "VolGridReader.h"

void VolGridReader::writeUnsignedShort(const void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(unsigned short);
	fwrite(p, size, nofItems, fp);
}


void VolGridReader::writeLong(const void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(long);
	fwrite(p, size, nofItems, fp);
}


void VolGridReader::writeFloat(const void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(float);
	fwrite(p, size, nofItems, fp);
}


void VolGridReader::writeInt(const void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(int);
	fwrite(p, size, nofItems, fp);
}


void VolGridReader::writeBool(const void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(bool);
	fwrite(p, size, nofItems, fp);
}


void VolGridReader::readUnsignedShort(void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(unsigned short);
	fread(p, size, nofItems, fp);
}

void VolGridReader::readLong(void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(long);
	fread(p, size, nofItems, fp);
}


void VolGridReader::readFloat(void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(float);
	fread(p, size, nofItems, fp);
}


void VolGridReader::readInt(void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(int);
	fread(p, size, nofItems, fp);
}


void VolGridReader::readBool(void *p, size_t nofItems, FILE *fp) {	
	size_t size = sizeof(bool);
	fread(p, size, nofItems, fp);
}


bool VolGridReader::save(FILE *fp) {

	if (fp == NULL) {
		return false;
	}
	
	long distanceValuesOrigin = -1;

	// write offsets for distance values
	this->writeLong(&distanceValuesOrigin, 1, fp);

	this->distanceValuesMap.clear();
	this->diffusionValuesMap.clear();
	this->distanceValuesSet.clear();

	// write octree parameters
	this->writeInt(&this->grid->maxActualDepth, 1, fp);
	this->writeFloat(&this->grid->center[0], 1, fp);
	this->writeFloat(&this->grid->center[1], 1, fp);
	this->writeFloat(&this->grid->center[2], 1, fp);
	this->writeFloat(&this->grid->length, 1, fp);
	this->writeInt(&this->grid->nLevels, 1, fp);
	this->writeInt(&this->grid->rootLevel, 1, fp);
	this->writeFloat(&this->grid->maxVal, 1, fp);

	// write volgrid parameters
	this->writeFloat(&this->grid->boundingBoxMin[0], 1, fp);
	this->writeFloat(&this->grid->boundingBoxMin[1], 1, fp);
	this->writeFloat(&this->grid->boundingBoxMin[2], 1, fp);
	this->writeFloat(&this->grid->boundingBoxMax[0], 1, fp);
	this->writeFloat(&this->grid->boundingBoxMax[1], 1, fp);
	this->writeFloat(&this->grid->boundingBoxMax[2], 1, fp);
	this->writeInt(&this->grid->maxDepth, 1, fp);
	this->writeFloat(&this->grid->narrowBandWidth, 1, fp);

	this->distanceValueCounter = 0;
	this->saveNode((VolGridNode *)this->grid->root);

	distanceValuesOrigin = ftell(fp);
	long size = this->distanceValuesSet.size();
	this->writeLong(&size, 1, fp);

	std::set<float *>::iterator iter = this->distanceValuesSet.begin();
	while (iter != this->distanceValuesSet.end()) {
		
		this->writeInt(&(*iter), 1, fp);
		this->writeFloat(&(**iter), 1, fp);
		iter++;

	}

	fseek(fp, 0, SEEK_SET);
	this->writeLong(&distanceValuesOrigin, 1, fp);
	

	fclose(fp);

	return true;
}

void VolGridReader::saveNode(VolGridNode *node) {

    int i, j;
	this->writeUnsignedShort(&node->nodeType, 1, fp);
	this->writeUnsignedShort(&node->xLocCode, 1, fp);
	this->writeUnsignedShort(&node->yLocCode, 1, fp);
	this->writeUnsignedShort(&node->zLocCode, 1, fp);
	this->writeUnsignedShort(&node->level, 1, fp);

	for (i = 0; i < 8; i++) {
		
		j = 0;
		this->distanceValuesSet.insert(node->distanceValues[i]);
		j = (int) node->distanceValues[i];


		this->writeInt(&j, 1, fp);
	}

	this->writeBool(&node->valid, 1, fp);
	this->writeBool(&node->boundaryNode, 1, fp);

    // recurse into children
	if ((VolGridNode *)node->children) {
		for (i = 0; i < 8; i++) {
			saveNode((VolGridNode *)node->children[i]);
		}
	}
}


bool VolGridReader::load(FILE *fp) {
	
	if (fp == NULL) {
		return false;
	}

	long distanceValuesOrigin = -1;
	long continueOrigin;

	// write offsets for distance and diffusion values
	this->readLong(&distanceValuesOrigin, 1, fp);

	if (distanceValuesOrigin == -1) {
		return false;
	}

	continueOrigin = ftell(fp);

	this->distanceValuesMap.clear();
	this->diffusionValuesMap.clear();
	this->distanceValuesSet.clear();

	fseek(fp, distanceValuesOrigin, SEEK_SET);
	long size;
	this->readLong(&size, 1, fp);

	for (long i = 0; i < size; i++) {
		int address;
		float *distanceValue = new float();
		float *diffusionValue = new float();
		this->readInt(&address, 1, fp);
		this->readFloat(&(*distanceValue), 1, fp);
		*diffusionValue = *distanceValue;
		this->distanceValuesMap.insert(std::map<int, float*>::value_type(address, distanceValue));
		this->diffusionValuesMap.insert(std::map<int, float*>::value_type(address, diffusionValue));
	}

	// rewind the file position
	fseek(fp, continueOrigin, SEEK_SET);

  	VolGridNode *node = new VolGridNode();
  	node->parent = NULL;
  	node->children = NULL;
	this->grid->root = node;
	
	// read octree parameters
	this->readInt(&this->grid->maxActualDepth, 1, fp);
	this->readFloat(&this->grid->center[0], 1, fp);
	this->readFloat(&this->grid->center[1], 1, fp);
	this->readFloat(&this->grid->center[2], 1, fp);
	this->readFloat(&this->grid->length, 1, fp);
	this->readInt(&this->grid->nLevels, 1, fp);
	this->readInt(&this->grid->rootLevel, 1, fp);
	this->readFloat(&this->grid->maxVal, 1, fp);

	// read volgrid parameters
	this->readFloat(&this->grid->boundingBoxMin[0], 1, fp);
	this->readFloat(&this->grid->boundingBoxMin[1], 1, fp);
	this->readFloat(&this->grid->boundingBoxMin[2], 1, fp);
	this->readFloat(&this->grid->boundingBoxMax[0], 1, fp);
	this->readFloat(&this->grid->boundingBoxMax[1], 1, fp);
	this->readFloat(&this->grid->boundingBoxMax[2], 1, fp);
	this->readInt(&this->grid->maxDepth, 1, fp);
	this->readFloat(&this->grid->narrowBandWidth, 1, fp);

	loadNode(node);

	return true;
}


void VolGridReader::loadNode(VolGridNode *node) {

	int i;
	int address;

	this->writeUnsignedShort(&node->nodeType, 1, fp);
	this->writeUnsignedShort(&node->xLocCode, 1, fp);
	this->writeUnsignedShort(&node->yLocCode, 1, fp);
	this->writeUnsignedShort(&node->zLocCode, 1, fp);
	this->writeUnsignedShort(&node->level, 1, fp);
	for (i = 0; i < 8; i++) {
		this->readInt(&address, 1, fp);
		std::map<int, float*>::iterator iter;
		iter = (this->distanceValuesMap.find(address));
		node->distanceValues[i] = (*iter).second;
		iter = (this->diffusionValuesMap.find(address));
		node->diffusionValues[i] = (*iter).second;
	}
	this->readBool(&node->valid, 1, fp);
	this->readBool(&node->boundaryNode, 1, fp);

	if (node->nodeType == INTERNAL) {
		node->children = (OcTreeNode **) new VolGridNode *[8];
		for (i = 0; i < 8; i++) {

			node->children[i] = new VolGridNode();
	        node->children[i]->parent = node;
			loadNode((VolGridNode *)node->children[i]);
		}
	}
	else if ((node->nodeType == LEAF) || (node->nodeType == EXTERNAL)) {
		node->children = NULL;
	}
	else {
		// something went wrong
		return;
	}
}
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
